import channels.layers
from asgiref.sync import async_to_sync
from collections import defaultdict
from django.db.models.signals import (pre_save, post_save,
                                      pre_delete, post_delete,
                                      post_migrate)
from django.db import transaction, models, ProgrammingError
from .mongo import MongoSignalWriter
from .redis import RedisSession, sync_get_tstamp
from .exceptions import RxDjangoBug


class SignalHandler:

    def __init__(self, channel_class):
        self.channel_class = channel_class
        self.meta = channel_class.meta
        self.name = channel_class.name
        self.state_model = channel_class._state_model
        self.wsrouter = channel_class._wsrouter
        self.mongo = MongoSignalWriter(channel_class)
        self.relay_map = defaultdict(list)

        self._setup = False

    def setup(self, app_config):
        if self._setup:
            return
        self._setup = True

        def init_cache_database(sender, **kwargs):
            self.mongo.init_database()
            RedisSession.init_database(self.channel_class)

        # Cache is deleted on every migrate
        post_migrate.connect(
            init_cache_database,
            sender=app_config,
            weak=False,
        )

        for model_layer in self.state_model.models():
            self._connect_layer(model_layer)

        if self.channel_class.meta.auto_update:
            self._connect_anchor_events()

    def _connect_layer(self, layer):
        """Register signals for models of this layer"""
        def relay_instance_optimistically(sender, instance, **kwargs):
            if sender is layer.model:
                tstamp = sync_get_tstamp()
                if not instance.id:
                    # Doesn't matter to optimistically send new object if
                    # referring instance is not updated
                    # This can be implemented by sending a patch on the parent,
                    # using only the id, without hitting the database
                    return

                serialized = layer.serialize_instance(instance, tstamp)
                serialized['_operation'] = 'update'
                serialized['_optimistic'] = True
                self._schedule(serialized, layer)

        def prepare_save(sender, instance, **kwargs):
            # Check if this instance has changed parent
            # If so we need to relay the old and new parent
            if not layer.reverse_acessor:
                return
            if kwargs.get('created', False):
                instance.__parent_updated = True
                return
            try:
                current = sender.objects.get(pk=instance.pk or instance.id)
            except sender.DoesNotExist:
                return # Should not happen, but who knows
            acessor = layer.reverse_acessor
            try:
                old_parent = getattr(current, acessor)
            except AttributeError as e:
                # This is a bug in RxDjango.
                # While we don't fix it, let's not break things
                return
            new_parent = getattr(instance, acessor)
            if new_parent != old_parent:
                instance.__parent_updated = True
                instance.__old_parent = old_parent

        def _relay_instance(_layer, instance, tstamp, operation, already_relayed=None):
            if not instance:
                return
            if already_relayed is None:
                already_relayed = set()
            if isinstance(instance, models.Model):
                instances = [instance]
            elif isinstance(instance, models.Manager):
                instances = instance.all()
            else:
                raise ProgrammingError()

            for _instance in instances:
                if operation == 'delete':
                    serialized = _layer.serialize_delete(_instance, tstamp)
                else:
                    serialized = _layer.serialize_instance(_instance, tstamp)
                serialized['_operation'] = operation
                key = f'{_layer.instance_type}:{_instance.id}'
                if key in already_relayed:
                    continue
                already_relayed.add(key)

                self._schedule(serialized, _layer)

                if operation == 'create':
                    # If instance is being created in this channel,
                    # then all related objects need to be scheduled
                    for attribute, child_layer in _layer.children.items():
                        child = getattr(_instance, attribute, None)
                        if child is None:
                            continue
                        elif isinstance(child, models.QuerySet):
                            children = child.all()
                        else:
                            children = [child]

                        for child in children:
                            try:
                                if child.id is None:
                                    continue
                            except AttributeError:
                                pass
                            _relay_instance(child_layer, child, tstamp, operation, already_relayed)

        def relay_instance(sender, instance, **kwargs):
            if sender is layer.model:
                tstamp = sync_get_tstamp()
                created = kwargs.get('created', None)
                operation = kwargs.get('_operation', 'update' if not created else 'create')
                if operation == 'create':
                    created = True
                _relay_instance(layer, instance, tstamp, operation)
                if not layer.origin or not layer.reverse_acessor:
                    return
                if created or getattr(instance, '__parent_updated', False):
                    parent = instance
                    for reverse_acessor in layer.reverse_acessor.split('.'):
                        parent = getattr(parent, reverse_acessor, None)
                    _relay_instance(layer.origin, parent, tstamp, 'update')
                    old_pa = getattr(instance, '__old_parent', None)
                    _relay_instance(layer.origin, old_pa, tstamp, 'update')

        def prepare_deletion(sender, instance, **kwargs):
            """Obtain anchors prior to deletion and store in instance"""
            if sender is layer.model:
                tstamp = sync_get_tstamp()
                serialized = layer.serialize_delete(instance, tstamp)
                anchors = layer.get_anchors(serialized)
                instance._anchors = list(anchors)
                instance._serialized = serialized

        def relay_delete_instance(sender, instance, **kwargs):
            if sender is layer.model:
                serialized = instance._serialized
                self._schedule(serialized, layer, instance._anchors)

        uid = '-'.join(['cache', self.name] + layer.instance_path)

        if layer.optimistic:
            pre_save.connect(
                relay_instance_optimistically,
                sender=layer.model,
                dispatch_uid=f'{uid}-optimistic',
                weak=False,
            )

        pre_save.connect(
            prepare_save,
            sender=layer.model,
            dispatch_uid=f'{uid}-prepare-save',
            weak=False,
        )

        post_save.connect(
            relay_instance,
            sender=layer.model,
            dispatch_uid=f'{uid}-save',
            weak=False,
        )

        pre_delete.connect(
            prepare_deletion,
            sender=layer.model,
            dispatch_uid=f'{uid}-prepare-deletion',
            weak=False,
        )

        post_delete.connect(
            relay_delete_instance,
            sender=layer.model,
            dispatch_uid=f'{uid}-delete',
            weak=False,
        )

        self.relay_map[layer.model].append(relay_instance)

    def _connect_anchor_events(self):
        """Add signal to broadcast creation and deletion of instances"""
        anchor_model = self.channel_class.meta.state.child.Meta.model
        channel_layer = channels.layers.get_channel_layer()

        def add_to_list(sender, instance, **kwargs):
            if sender is not anchor_model or not kwargs.get('created', False):
                return
            async_to_sync(channel_layer.group_send)(
                self.channel_class._anchor_events_channel,
                {
                    'type': 'instances.list.add',
                    'instance_id': instance.id,
                },
            )
        def remove_from_list(sender, instance, **kwargs):
            if sender is not anchor_model:
                return
            async_to_sync(channel_layer.group_send)(
                self.channel_class._anchor_events_channel,
                {
                    'type': 'instances.list.remove',
                    'instance_id': instance._serialized['id'],
                },
            )

        post_save.connect(
            add_to_list,
            sender=anchor_model,
            dispatch_uid=f'{anchor_model.__name__}-list-add',
            weak=False,
        )
        post_delete.connect(
            remove_from_list,
            sender=anchor_model,
            dispatch_uid=f'{anchor_model.__name__}-list-remove',
            weak=False,
        )

    def _schedule(self, serialized, state_model, anchors=None):
        if serialized['id'] is None and serialized['_operation'] == 'create':
            raise RxDjangoBug('Saving instance without id causes data leakage. '
                              'Check stack trace to fix this bug.')
        if transaction.get_autocommit():
            self._relay(serialized, state_model, anchors)
            return

        transaction.on_commit(
            lambda: self._relay(serialized, state_model, anchors)
        )

    def _relay(self, serialized, state_model, anchors):
        """Send one update for both cache and connected clients"""
        user_id = serialized.get('_user_key', None)
        payload = [serialized]
        if anchors is None:
            anchors = state_model.get_anchors(serialized)
        for anchor in anchors:
            deltas = self.mongo.write_instances(anchor.id, payload)
            if deltas:
                self.wsrouter.sync_dispatch(deltas, anchor.id, user_id)

    def broadcast_instance(self, anchor_id, instance, operation='update'):
        kwargs = {
            '_operation': operation
        }
        sender = instance.__class__
        for relay_instance in self.relay_map[sender]:
            relay_instance(sender, instance, **kwargs)



import json
from datetime import datetime
from pytz import utc
from django.utils import timezone
from django.db import models
from django.db.models.query import QuerySet
from asgiref.sync import sync_to_async, async_to_sync
from channels.db import database_sync_to_async
from channels.generic.websocket import AsyncWebsocketConsumer
from rest_framework.authtoken.models import Token
from .state_loader import StateLoader
from .actions import execute_action
from .exceptions import UnauthorizedError, ForbiddenError, AnchorDoesNotExist
from rxdjango.serialize import json_dumps


class StateConsumer(AsyncWebsocketConsumer):
    """
    WebSocket consumer that manages user authentication, session management,
    and real-time data relay to clients. A subclass of StateConsumer will be
    dinamically created by ContextChannel.as_asgi().
    """
    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.channel = None
        self.user = None
        self.token = None
        self.anchor_ids = None
        self.wsrouter = None
        self.tstamp = None
        self.session = None

    async def connect(self):
        """Accept any connection and just wait for a token."""
        await self.accept()

    async def receive(self, text_data):
        try:
            data = json.loads(text_data)
        except json.JSONDecodeError:
            await self.disconnect()
            raise

        if self.user:
            await self.receive_action(data)
        else:
            await self.receive_authentication(data)

    async def receive_authentication(self, text_data):
        # If user is not logged, we expect credentials
        # data = json.loads(text_data)
        data = text_data
        token = data.get('token', None)
        last_update  = data.get('last_update', None)

        user = None
        try:
            user = await self.authenticate(token=token)
        except UnauthorizedError:
            await self.send_connection_status(401, 'error/unauthorized')
        except ForbiddenError:
            await self.send_connection_status(403, 'error/forbidden')

        if user:
            try:
                await self.start(user, last_update)
            except AnchorDoesNotExist:
                await self.send_connection_status(404, 'error/not-found')

        else:
            await self.close()

    async def start(self, user, tstamp):
        kwargs = self.scope['url_route']['kwargs']
        self.user = user
        self.channel = self.context_channel_class(user, **kwargs)
        await self.channel.initialize_anchors()
        self.channel._consumer = self
        self.user_id = self.user.id
        self.anchor_ids = self.channel.anchor_ids
        self.wsrouter = self.channel._wsrouter

        await self.send_connection_status(200)

        for anchor_id in self.anchor_ids:
            await self.connect_anchor(anchor_id)

        if self.channel.meta.auto_update:
            await self.channel_layer.group_add(
                self.channel._anchor_events_channel,
                self.channel_name,
            )

        await self.channel.on_connect(tstamp)

        for anchor_id in self.anchor_ids:
            await self._load_state(anchor_id)

    # Called by channel layers
    async def instances_list_add(self, event):
        instance_id = event['instance_id']
        if await self.channel.is_visible(instance_id):
            await self.channel.add_instance(instance_id, at_beginning=True)

    async def instances_list_remove(self, event):
        instance_id = event['instance_id']
        await self.channel.remove_instance(instance_id)

    async def connect_anchor(self, anchor_id):
        await self.wsrouter.connect(
            self.channel_layer,
            self.channel_name,
            anchor_id,
            self.user_id,
        )

    async def disconnect_anchor(self, anchor_id):
        await self.wsrouter.disconnect(
            self.channel_layer,
            self.channel_name,
            anchor_id,
            self.user_id,
        )

    async def _load_state(self, anchor_id):
        async with StateLoader(self.channel, anchor_id) as loader:
            async for instances in loader.list_instances():
                if instances:
                    data = json_dumps(instances)
                    await self.send(text_data=data)

            self.tstamp = loader.tstamp
            await self.send(text_data=self.end_of_data)

    @property
    def end_of_data(self):
        return json_dumps([{
            '_instance_type': '',
            '_tstamp': self.tstamp,
            '_operation': 'end_initial_state',
            'id': 0,
        }])

    @database_sync_to_async
    def serialized_data(self, Serializer, data):
        return Serializer(data).data

    async def disconnect(self, close_code=None):
        if not self.channel:
            return

        for anchor_id in self.anchor_ids:
            await self.wsrouter.disconnect(
                self.channel_layer,
                self.channel_name,
                anchor_id,
                self.user_id,
            )

        await self.channel.on_disconnect()

    @database_sync_to_async
    def authenticate(self, token):
        try:
            token = Token.objects.get(key=token)
        except Token.DoesNotExist:
            raise UnauthorizedError('error/unauthorized')
            return
        self.token = token.key

        kwargs = self.scope['url_route']['kwargs']
        if not self.context_channel_class.has_permission(
                token.user,
                **kwargs,
        ):
            raise ForbiddenError('error/forbidden')
            return

        return token.user

    async def relay(self, payload):
        payload = payload['payload']
        await self.send(text_data=json_dumps(payload))

    async def receive_action(self, action):
        call_id = action['callId']
        method_name = action.pop('action')
        params = action.pop('params')

        try:
            action['result'] = await execute_action(self.channel, method_name, params)
            await self.send(text_data=json.dumps(action))
        except Exception as e:
            action['error'] = 'Error'
            await self.send(text_data=json.dumps(action))
            raise

    async def send_connection_status(self, status_code, error=None):
        data = {}
        data['status_code'] = status_code
        if error:
            data['error'] = error
        text_data = json_dumps(data)
        close = error != None
        await self.send(text_data=text_data, close=close)

    async def prepend_anchor_id(self, anchor_id):
        data = {
            'prependAnchor': anchor_id,
        }
        text_data = json_dumps(data)
        await self.send(text_data=text_data)



import asyncio
from asgiref.sync import sync_to_async
from channels.db import database_sync_to_async
from django.conf import settings
from .redis import RedisStateSession
from .mongo import MongoStateSession
from .exceptions import AnchorDoesNotExist

if settings.DEBUG or settings.TESTING:
    def mark(instances, cache_state):
        if instances:
            instances[0]['_cache_state'] = cache_state
        return instances
else:
    def mark(instances, cache_state):
        return instances


class StateLoader:
    """Manages the cache synchronization process using Redis.
    Cache has 4 possible states, and each time a client access it, a transition
    may occur. These are:

    - 0 COLD: No cache, need to get from db
    - 1 HEATING: Cache is being built, peers can follow at redis
    - 2 HOT: Cache is available, everyone gets from mongo
    - 3 COOLING: Cache is being deleted, first peer can re-heat from redis

    Attributes:
        name (str): The name of the cache.
        anchor_id (int): The anchor ID associated with the cache.
        base_key (str): The base key used for all Redis keys related to this anchor ID.
    """

    def __init__(self, channel, anchor_id):
        self.channel = channel
        self.state_model = channel._state_model

        self.redis = RedisStateSession(channel, anchor_id)
        self.mongo = MongoStateSession(channel, anchor_id)

        self.anchor_id = anchor_id
        self.user_id = channel.user_id

        self.cache_state = None
        self.tstamp = None

    async def __aenter__(self):
        self.cache_state = await self.redis.load()
        self.tstamp = self.redis.tstamp
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.redis.end(exc_type is None)

        if exc_type is not None:
            raise exc

    def _user_filter(self, instances):
        """Filter out other user's instances.
        During COLD, HEATING and COOLING loader has to deal with all users
        """
        return [
            i for i in instances if i.pop('_user_key') in (self.user_id, None)
        ]

    async def _list_instances_cold(self):
        """In COLD state, fetch all instances from database,
        write to mongo and redis and send instances filtered by user
        """
        anchor = await self._get_anchor_from_db()
        iterator = self.state_model.serialize_state(anchor, self.tstamp)

        @database_sync_to_async
        def next_instances():
            return next(iterator, None)

        while True:
            instances = await next_instances()
            if instances is None:
                break

            await asyncio.gather(
                self.mongo.write_instances(instances),
                self.redis.write_instances(instances),
            )

            instances = self._user_filter(instances)
            if instances:
                yield mark(instances, 'cold')

    @database_sync_to_async
    def _get_anchor_from_db(self):
        Anchor = self.state_model.model
        try:
            return Anchor.objects.get(id=self.anchor_id)
        except Anchor.DoesNotExist:
            raise AnchorDoesNotExist(
                f'{Anchor.__name__} id {self.anchor_id} does not exist'
            )

    async def _list_instances_heating(self):
        """In HEATING state, fetch all instances from redis and
        send instances filtered by user.
        """
        async for instances in self.redis.list_instances():
            instances = self._user_filter(instances)
            yield mark(instances, 'heating')

    async def _list_instances_hot(self):
        """In HOT state, fetch filtered instances from mongo and send"""
        async for instances in self.mongo.list_instances(self.user_id):
            yield mark(instances, 'hot')

    async def _list_instances_cooling(self):
        """In COOLING state, fetch all instances from redis, write to mongo
        and send instances filtered by user.
        """
        async for instances in self.redis.list_instances():
            await self.mongo.write_instances(instances)
            instances = self._filter(instances)
            yield mark(instances, 'cooling')

    router = [
        _list_instances_cold,
        _list_instances_heating,
        _list_instances_hot,
        _list_instances_cooling,
    ]

    async def list_instances(self):
        """All instances that should go to a user"""
        if self.cache_state is None:
            raise Exception("Generator must be called inside with block")

        list_method = self.router[self.cache_state]

        async for instances in list_method(self):
            yield instances



import importlib
from django.apps import apps, AppConfig


class ReactFrameworkConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'rxdjango'

    def ready(self):
        """Discover and register ContextChannel subclasses within Django apps."""
        from . import channels

        for app_config in apps.get_app_configs():
            try:
                # Attempt to import the channels.py module from the app
                channels_module = importlib.import_module(f"{app_config.name}.channels")

                # Check for subclasses of ContextChannel in the module
                for attr_name in dir(channels_module):
                    attr = getattr(channels_module, attr_name)
                    # Register the subclass in the global dictionary
                    if not isinstance(attr, type) or \
                       not issubclass(attr, channels.ContextChannel) or \
                       attr.Meta.abstract:
                        continue
                    attr._signal_handler.setup(app_config)

            except ImportError:
                # channels.py not found in the app, so just continue
                pass



from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from rxdjango.ts.interfaces import create_app_interfaces
from rxdjango.ts.channels import create_app_channels


def make_sdk(apply_changes=True, quiet=False, force=False):
    print("Generating RxDjango SDK")
    check()

    def log(msg):
        if not quiet:
            print(msg)

    models = apps.get_models()
    installed_apps = list(set([ x.__module__.split('.')[0] for x in models]))

    changed = False

    for app in installed_apps:
        diff = create_app_interfaces(app, apply_changes, force)
        if diff:
            changed = True
            log(diff)
        diff = create_app_channels(app, apply_changes, force)
        if diff:
            changed = True
            log(diff)

    return changed


def check():
    if not getattr(settings, 'RX_FRONTEND_DIR', None):
        raise ImproperlyConfigured(
            "settings.RX_FRONTEND_DIR is not set. Configure it with a folder "
            "inside your react application."
        )



import sys
import asyncio
from django.core.management.base import BaseCommand
from rxdjango.sdk import make_sdk
from rxdjango.websocket_router import send_system_message


class Command(BaseCommand):
    help="Broadcast a system message to all connected clients"

    def add_arguments(self, parser):
        parser.add_argument('source',
                            type=str,
                            help=(
                                'Source of the broadcast message'
                                '(currently only "maintenance" is supported)'
                            ))

        parser.add_argument('message',
                            type=str,
                            help='Verbose message to users',
                            )

    def handle(self, *args, **options):
        asyncio.run(
            send_system_message(
                options['source'],
                options.get('message'),
            )
        )



from setuptools import setup, find_packages
from os import path

cur_dir = path.abspath(path.dirname(__file__))

VERSION = '0.0.22'

setup(
    name="rxdjango",
    author="Luis Fagundes",
    author_email="lhfagundes@gmail.com",
    version=VERSION,
    packages=find_packages(),
    license="LICENSE.md",
    install_requires=[
        'Django>=4.2',
        'motor>=3.3',
        'channels>=4',
        'channels-redis>=4.1',
        'djangorestframework>=3',
        'daphne>=4.1.0',
        'pytz',
    ],
    url="https://github.com/CDIGlobalTrack/rxdjango",
    include_package_data=True,
    python_requires=">=3.10"
)



from daphne.management.commands.runserver import Command as RunserverCommand
from rxdjango.sdk import make_sdk

class Command(RunserverCommand):

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument('--makefrontend', action='store_true',
                            help="Build the frontend SDK files")

    def inner_run(self, *args, **kwargs):
        if kwargs['makefrontend']:
            make_sdk()
        super().inner_run(*args, **kwargs)



import asyncio
import typing
import inspect
from collections import defaultdict
from .exceptions import ForbiddenError, ActionNotAsync

__actions = set()


def action(method):
    """Methods in a ContextChannel decorated with @action will be
    implemented in the frontend counterpart ContextChannel class,
    so that the method can be called directly from the frontend.
    """
    if not asyncio.iscoroutinefunction(method):
        raise ActionNotAsync(f'@action decorator requires "{method.__name__}" to be async')
    wrapped = method
    while getattr(wrapped, '__wrapped__', None):
        wrapped = wrapped.__wrapped__
    __actions.add(wrapped)
    return method


def list_actions(channel):
    """List all decorated methods in this django deployment"""
    for method in channel.__dict__.values():
        if method in __actions:
            yield method


async def execute_action(channel, method_name, params):
    method = getattr(channel, method_name, None)
    _verify_method(method)
    return await method(*params)


def _verify_method(method):
    if not method:
        raise ForbiddenError
    if getattr(method, '__func__', None):
        method = method.__func__
    while getattr(method, '__wrapped__', None):
        method = method.__wrapped__
    if not method in __actions:
        raise ForbiddenError




import os
import typing
import importlib
from django.utils import timezone
from django.db.models.fields import related_descriptors
from django.conf import settings
from rest_framework import serializers, relations, fields
from . import ts_exported, header, interface_name, get_ts_type, diff

def create_app_interfaces(app, apply_changes=True, force=False):
    serializer_module_name = f'{app}.serializers'
    serializer_path = serializer_module_name.replace('.', '/') + '.py'

    ts_file_path = os.path.join(settings.RX_FRONTEND_DIR, f'{app}/{app}.interfaces.d.ts')
    py_mtime = None

    if os.path.exists(serializer_path):
        py_mtime = os.path.getmtime(serializer_path)

    existing = []

    if os.path.exists(ts_file_path):
        if not force and py_mtime == os.path.getmtime(ts_file_path):
            return
        with open(ts_file_path, 'r') as file:
            existing = file.read().split('\n')

    try:
        module = importlib.import_module(serializer_module_name)
    except ModuleNotFoundError:
        return

    serializers = get_serializers(module)

    try:
        module_serializers = serializers.pop(app)
    except KeyError:
        return

    code = header(
        app,
        f'Based on {app}/serializers.py',
    )

    initial_length = len(code)

    code.append(f"import {{ InstanceType }} from '@rxdjango/react';\n")

    imports = []
    for external_app, dependencies in serializers.items():
        dependencies = [dep for dep in dependencies if ts_exported(dep)]
        if not dependencies:
            continue
        interfaces = [interface_name(d) for d in dependencies]
        imports.append(''.join([
            'import { ',
            ', '.join(sorted(interfaces)),
            ' } from ',
            f"'../{external_app}/{external_app}.interfaces';",
        ]))

    code += sorted(imports)

    code.append('')  # line break

    for Serializer in module_serializers:
        if ts_exported(Serializer):
            code.append(serialize_type(Serializer))

    if len(code) == initial_length + 2:
        return

    content = '\n'.join(code)

    if content.split('\n')[2:] == existing[2:]:
        return

    difference = diff(existing, content.split('\n'), ts_file_path)
    if not apply_changes:
        return difference

    try:
        with open(ts_file_path, 'w') as fh:
            fh.write(content)
    except FileNotFoundError:
        os.makedirs(os.path.dirname(ts_file_path), exist_ok=True)
        with open(ts_file_path, 'w') as fh:
            fh.write(content)

    if py_mtime:
        os.utime(ts_file_path, (py_mtime, py_mtime))

    return difference


def get_serializers(module):
    apps = {}

    for klass in _get_serializer_classes(module):
        app = klass.__module__.split('.')[0]
        try:
            apps[app].append(klass)
        except KeyError:
            apps[app] = [klass]
    return apps

def _get_serializer_classes(module):
    for name, klass in module.__dict__.items():
        try:
            if issubclass(klass, serializers.Serializer):
                yield klass
        except TypeError:
            pass

mappings = {
    serializers.BooleanField: 'boolean',
    serializers.CharField: 'string',
    serializers.EmailField: 'string',
    serializers.RegexField: 'string',
    serializers.SlugField: 'string',
    serializers.URLField: 'string',
    serializers.UUIDField: 'string',
    serializers.FilePathField: 'string',
    serializers.IPAddressField: 'string',
    serializers.IntegerField: 'number',
    serializers.FloatField: 'number',
    serializers.DecimalField: 'number',
    serializers.DateTimeField: 'string',
    serializers.DateField: 'string',
    serializers.TimeField: 'string',
    serializers.DurationField: 'string',
    serializers.DictField: 'Map',
    relations.PrimaryKeyRelatedField: 'number',
}


def __process_field(field_name, field, serializer):
    '''
    Generates and returns a tuple representing the Typescript field name and Type.
    '''
    if hasattr(field, 'child'):
        is_many = True
        field_type = type(field.child)
    elif hasattr(field, 'child_relation'):
        is_many = True
        field_type = type(field.child_relation)
    else:
        is_many = False
        field_type = type(field)

    ts_type = mappings.get(field_type)

    if not ts_type:
        if hasattr(field, 'choices'):
            ts_type = __map_choices_to_union(field_type, field.choices)
        elif field_type is fields.ReadOnlyField:
            model_field = getattr(serializer.Meta.model, field_name)
            if isinstance(model_field,
                          related_descriptors.ForeignKeyDeferredAttribute):
                ts_type = 'number'
            else:
                func = model_field
                if type(model_field) is property:
                    func = func.fget
                ts_type = __get_function_type(func)

        else:
            if field_type.__name__.endswith('Serializer'):
                ts_type = interface_name(field_type)
            else:
                ts_type = 'any'


    if is_many:
        ts_type += '[]'

    return (field_name, ts_type)

def __get_function_type(func):
    hints = typing.get_type_hints(func)
    try:
        ftype = hints['return']
    except KeyError:
        return 'any'

    return get_ts_type(ftype)

def __map_choices_to_union(field_type, choices):
    '''
    Generates and returns a TS union type for all values in the provided choices OrderedDict
    '''
    if not choices:
        return 'any'

    return ' | '.join(f'"{key}"' if type(key) == str else str(key) for key in choices.keys())


def serialize_type(serializer):
    '''
    Generates and returns a Typescript Interface by iterating
    through the serializer fields of the DRF Serializer class
    passed in as a parameter, and mapping them to the appropriate Typescript
    data type.
    '''
    name = serializer.__name__.replace('Serializer', 'Type')
    fields = []
    if hasattr(serializer, 'get_fields'):
        instance = serializer()
        fields = instance.get_fields().items()
    else:
        try:
            fields = serializer._declared_fields.items()
        except AttributeError:
            return
    ts_fields = []
    for key, value in fields:
        property, type = __process_field(key, value, serializer)

        if key != 'id':
            if value.required:
                property = property + "?"

            if value.allow_null and not value.read_only:
                type = type + " | null"

        ts_fields.append(f"    {property}: {type};")
    collapsed_fields = '\n'.join(ts_fields)
    return f'export interface {name} extends InstanceType {{\n{collapsed_fields}\n}}\n'
